//
// K2Client.cpp : a very basic C++ test app for K2 AppService
//
// Below is a very basic AppServer client. It's purpose is to 
// demonstrate the ways of connecting to and using the AppService.
// It is up to the developer to add error checking, try/catch mechanisms
// and other safe guards to their programs.
//
// This file contains both a K2Class and a basic main function. Both were
// intentionally left in one file only to show the full functionality in 
// one place.

#include <afx.h>
#include <time.h>
#include <conio.h>

#import <mscorlib.tlb>

#include "GrassValley.Mseries.AppServerLog.tlh"
#include "GrassValley.Mseries.ConfigMgr.tlh"
#include "GrassValley.Mseries.DVCapture.tlh"
#include "GrassValley.Mseries.Editor.tlh"
#include "GrassValley.Mseries.MediaMgr.tlh"
#include "GrassValley.Mseries.Status.tlh"
#include "GrassValley.Mseries.TransferQueue.tlh"
#include "GrassValley.Mseries.ChanStatus.tlh"
#include "GrassValley.Mseries.Control.tlh"
#include "GrassValley.Mseries.AppServer.tlh"

// Class: K2Client
// This is a basic class that encapsulates connecting to 
// and calling functions on a K2 AppServ.
class K2Client
{
	// member variables
	private:	
	HRESULT hr;
	_bstr_t m_bstrHost;
	_bstr_t m_bstrSuiteName;
	_bstr_t m_bstrAppName;
	_bstr_t m_bstrUsername;
	_bstr_t m_bstrPassword;
	_bstr_t m_bstrDomain;
	_bstr_t m_bstrChannel;

	IAppServerMgrProxyPtr spAppServerMgrProxy;
	IAppServerPtr spAppServer;	

	public:
	// K2Client constructor
    K2Client(char* appname, char* host) 
	{
		// setup some useful strings
		m_bstrAppName = _bstr_t(appname);
		m_bstrHost = _bstr_t(host);
		m_bstrSuiteName = _bstr_t("K2ClientTestSuite");
		m_bstrUsername = _bstr_t("Administrator");
		m_bstrPassword = _bstr_t("adminK2");
		m_bstrDomain = _bstr_t("");
		m_bstrChannel = _bstr_t("C1");
	}

	// Connect to a K2 AppService
	bool Connect()
	{
		// Initialize the COM library
		printf("coinit COM...\n");
		CoInitialize(NULL);

		// get a connection to the local AppSrvMgr proxy
		printf("create an AppServerMgrProxy...\n");
		IAppServerMgrProxyPtr spTemp(__uuidof(AppServerMgrProxy));
		spAppServerMgrProxy = spTemp;	

		// using the proxy, connect to the AppSrvMgr
		printf("set host...\n");
		hr = spAppServerMgrProxy->SetHost(m_bstrHost);

		// set the user credentials
		VARIANT_BOOL encrypted(FALSE);
		hr = spAppServerMgrProxy->SetUserCredentials(m_bstrUsername, m_bstrPassword, 
			m_bstrDomain, encrypted);

		// connect
		printf("connect to AppServerMgr...\n");
		VARIANT_BOOL retVal;
		hr = spAppServerMgrProxy->Connect(&retVal);
		//printf("connect hresult = 0x%x, retval = %d\n", hr, retVal);
		if ( !retVal )
		{
			printf("ERROR - couldn't connect to the AppServerMgrr\n");
			return false;
		}
		return true;
	}

	bool CreateAppServer()
	{
		// make sure we have a valid AppServerMgr
		if ( NULL == spAppServerMgrProxy )
			return false;

		// Create an AppServer. When you ask for an AppServer, the AppServerMgr
		// will take a look at the suite name. If an AppServer has been created
		// for that name, it will be returned. If not, a new AppServer will be
		// created and returned. The "isNewConnection" variable will return
		// that information.

		printf("create AppServer...\n");
		short nNewConnection;
		hr = spAppServerMgrProxy->CreateAppServer(m_bstrSuiteName, m_bstrAppName, &nNewConnection, &spAppServer);

		if (spAppServer == NULL)
		{
			printf("invalid IAppServer returned...\n");
			return false;
		}

		return true;
	}

	void CheckSystemStatus()
	{
		// make sure the remote system is up and running first
		BSTR systemMode;
		BootStage bootStage;
		BootCode bootCode;
		printf("check system status...\n");
		spAppServerMgrProxy->GetSystemStatus(&systemMode, &bootStage, &bootCode);
		CString sSystemMode = systemMode;
		printf("SystemMode: %s  BootStage: %d  BootCode: %d\n\n", sSystemMode, bootStage, bootCode);

		if ( BootStage_UP_AND_RUNNING != bootStage || BootCode_NORMAL != bootCode)
			printf("ERROR - system not running in normal state!\n");
	}

	void GetChannelOwner()
	{
		_variant_t varChan("C1");
		BSTR bstrMachine;
		BSTR bstrSuite;
		BSTR bstrUser;
		BSTR bstrApp;

		printf("get channel owner...\n");
		spAppServerMgrProxy->GetChannelOwnerInfo(varChan, &bstrMachine, &bstrSuite, &bstrUser, &bstrApp);

		CString sMachine = bstrMachine;
		CString sSuite = bstrSuite;
		CString sUser = bstrUser;
		CString sApp = bstrApp;

		printf("machine = %s suite = %s user = %s app = %s\n", sMachine, sSuite, sUser, sApp);
	}

	void LogTestMessage()
	{
		printf("get log...\n");
		IAppServerLogPtr spLog;
		hr = spAppServer->CreateLog(m_bstrAppName, &spLog);
		//printf("create log hresult = 0x%x\n", hr);

		printf("write test message to log...\n\n");
		_bstr_t testmsg("test message");
		hr = spLog->LogMsg(EntryPtMsgType_eWarningMsg, 0, testmsg);
	}

	void PlayClip()
	{
		short nIsNewController;
		ISimpleControllerPtr spController;
		_variant_t varChanNum(m_bstrChannel);
		
		printf("create controller\n");
		hr = spAppServer->CreateController(m_bstrAppName, varChanNum, 
			&nIsNewController, &spController);
		//printf("CreateController hresult = 0x%x\n", hr);

		printf("load clip\n");
		ISimplePlayerRecorderPtr spPlayer = (ISimplePlayerRecorderPtr) spController;
		hr = spPlayer->Load(_bstr_t("edl/cmf//local/V:/default/Clip"));
		//printf("Load hresult = 0x%x\n", hr);

		printf("play\n");
		hr = spPlayer->Play();
		Sleep(3000);

		printf("eject\n");
		hr = spPlayer->Eject();

		printf("close channel\n\n");
		hr = spController->CloseChannel();
	}

	void RecordClip()
	{
		short nIsNewController;
		ISimpleControllerPtr spController;
		_variant_t varChanNum(m_bstrChannel);
		
		printf("create controller\n");
		hr = spAppServer->CreateController(m_bstrAppName, varChanNum, 
			&nIsNewController, &spController);
		//printf("CreateController hresult = 0x%x\n", hr);

		spController->SetCurrentBin(_bstr_t("V:"), _bstr_t("default"));
		BSTR bstrNewClip;
		hr = spController->GenerateUniqueName(_bstr_t("test"), _bstr_t("_"), &bstrNewClip);

		ISimplePlayerRecorderPtr spRecorder = (ISimplePlayerRecorderPtr) spController;
		spRecorder->New(bstrNewClip);

		spRecorder->Record();
		Sleep(3000);
		spRecorder->Stop();

		printf("eject\n");
		hr = spRecorder->Eject();

		printf("close channel\n\n");
		hr = spController->CloseChannel();
	}

	void TransferClip()
	{
		// create a transfer queue connection
		IXfrQueuePtr spXfrQueue;
		hr = spAppServer->CreateXfrQueue(m_bstrAppName, &spXfrQueue);

		// transfer a clip
		long token;
		spXfrQueue->AddXfrItem(
			_bstr_t("MX-ABC/explodedFile/V:/default/Clip"),
			_bstr_t("MX-DEF/explodedFile/V:/default/Clip"),
			_bstr_t("Send to Stream"),
			&token);

		// query for a property
		VARIANT data;
		spXfrQueue->GetXfrItemProperty(token, _bstr_t("Percentage"), &data);
		long nPercentage = data.lVal;
		printf("percentage = %d\n", nPercentage);

		// get xml data about the transfer item
		BSTR xmlData;
		spXfrQueue->GetXfrItemXml(token, &xmlData);

		CString sXmlData = xmlData;
		printf("xmlData = %s\n\n", sXmlData);
	}

	void GetChanStatusProperties()
	{
		IChanStatusPtr spChanStatus;
		_variant_t varChan(m_bstrChannel);
		hr = spAppServer->CreateChanStatus(m_bstrAppName, varChan, &spChanStatus);
		
		VARIANT varResult;
		// get a single property
		hr = spChanStatus->GetStatus(_bstr_t("timeremainingstr"), &varResult);

		CString sTimeRemaining = varResult.bstrVal;		
		printf("TimeRemaining: %s\n", sTimeRemaining);


		// get multiple properties
		hr = spChanStatus->GetStatus(_bstr_t("ownerName+assetname+videoFormatStr"),
			&varResult);

		SAFEARRAY *pArray = varResult.parray;
		VARIANT *pVar = NULL;
		long arrIndex[1];

		// get the first property
		arrIndex[0] = 0;
        SafeArrayPtrOfIndex(pArray, arrIndex, (void **)&pVar);
		CString sOwnerName = pVar->bstrVal;		

		// get the second property
		arrIndex[0] = 1;
        SafeArrayPtrOfIndex(pArray, arrIndex, (void **)&pVar);
		CString sAssetName = pVar->bstrVal;		

		// get the third property
		arrIndex[0] = 2;
        SafeArrayPtrOfIndex(pArray, arrIndex, (void **)&pVar);
		CString sVideoFormat = pVar->bstrVal;		

		printf("OwnerName: %s AssetName: %s VideoFormat: %s\n\n", sOwnerName, sAssetName, sVideoFormat);
	}

	void GetAssetProperties()
	{
		IMediaMgrPtr spMediaMgr;
		hr = spAppServer->CreateMediaMgr(m_bstrAppName, &spMediaMgr);
		
		VARIANT varResult;
		// get a single property
		hr = spMediaMgr->GetProperty(_bstr_t("edl/cmf//local/V:/default/Clip"), 
			_bstr_t("assetType"), &varResult);

		CString sAssetType = varResult.bstrVal;		
		printf("asset type: %s\n", sAssetType);


		// get multiple properties
		hr = spMediaMgr->GetProperty(_bstr_t("edl/cmf//local/V:/default/Clip"), 
			_bstr_t("LengthStr+videoCompressionType+videoFormatStr"), &varResult);

		SAFEARRAY *pArray = varResult.parray;
		VARIANT *pVar = NULL;
		long arrIndex[1];

		// get the first property
		arrIndex[0] = 0;
        SafeArrayPtrOfIndex(pArray, arrIndex, (void **)&pVar);
		CString sLength = pVar->bstrVal;		

		// get the second property
		arrIndex[0] = 1;
        SafeArrayPtrOfIndex(pArray, arrIndex, (void **)&pVar);
		CString sVideoCompressionType = pVar->bstrVal;		

		// get the third property
		arrIndex[0] = 2;
        SafeArrayPtrOfIndex(pArray, arrIndex, (void **)&pVar);
		CString sVideoFormat = pVar->bstrVal;		

		printf("Length: %s VideoCompressionType: %s VideoFormat: %s\n\n", sLength, sVideoCompressionType, sVideoFormat);
	}

	void GetXmlChanges()
	{
		IMediaMgrPtr spMediaMgr;
		hr = spAppServer->CreateMediaMgr(m_bstrAppName, &spMediaMgr);

		long count = 0;
		BSTR bstrXmlData = NULL;		// must set this to NULL!

		spMediaMgr->GetXmlChanges(1000000, &count, &bstrXmlData);

		CString sXmlData = bstrXmlData;
		printf("GetXmlChanges returned:\n%s\n\n", sXmlData);
	}

	void CreatePlaylist()
	{
		short nIsNewController;
		ISimpleControllerPtr spController;
		_variant_t varChanNum(m_bstrChannel);
		
		// create the controller and cast it to different interfaces
		printf("create controller\n");
		hr = spAppServer->CreateController(m_bstrAppName, varChanNum, 
			&nIsNewController, &spController);
		//printf("CreateController hresult = 0x%x\n", hr);

		ISimplePlayerRecorderPtr spPlayer = (ISimplePlayerRecorderPtr) spController;
		ITwoHeadPlayerRecorderPtr spTwoHeadPlayer = (ITwoHeadPlayerRecorderPtr) spController;

		// set loop play mode to false
		VARIANT_BOOL loopmode(FALSE);
		printf("set loop play mode to false\n");
		hr = spPlayer->put_LoopPlayMode(loopmode);
		
		// enable list pauses
		printf("enable list pauses\n");
		hr = spController->SetChannelProperty(_bstr_t("listpauses"), _variant_t("true"));

		// enable list repeats
		printf("enable list repeats\n");
		hr = spController->SetChannelProperty(_bstr_t("listrepeats"), _variant_t("true"));
		
		// set current bin
		printf("set current bin\n");
		hr = spController->SetCurrentBin(_bstr_t("V:"), _bstr_t("default"));
			
		// generate a unique name for the playlist (i.e. playlist, playlist_0, playlist_1, etc.)
		printf("generate unique playlist name\n");
		BSTR playlist;
		hr = spController->GenerateUniqueName (_bstr_t("playlist"), _bstr_t("_"), &playlist);

		// eject the play first to make sure we're clean
		printf("eject channel to make sure we're clean\n");
		hr = spPlayer->Eject();

		// create the new playlist (which also loads it)
		printf("create new playlist\n");
		hr = spPlayer->New(playlist);

		// get an editor and cast it to an event editor
		printf("create an event editor\n");
		ISimpleEditorPtr spEditor;
		hr = spPlayer->GetEditor(&spEditor);
		IEventEditorPtr spEventEditor = (IEventEditorPtr) spEditor;

		// insert the first section
		printf("insert section\n");
		BSTR section1;
		hr = spEventEditor->InsertSection(_bstr_t(""), &section1);

		// set the default starting timecode (this is what AppCenter does).
		printf("set starting timecode\n");
		spEditor->SetStartingTimecode(_variant_t("01:00:00,00"));

		// insert several play events and get properties back
		printf("insert play event\n");
		BSTR editID_A;
		VARIANT varResult;
		hr = spEventEditor->InsertPlayEvent(_bstr_t("edl/cmf//local/V:/default/ClipA"), section1, 0, _bstr_t(""), &editID_A);
		printf("get event properties\n");
		hr = spEventEditor->GetEventProperty (editID_A, _bstr_t("name+startpos+duration"), &varResult);			
		// your code to do something with event properties

		printf("insert play event\n");
		BSTR editID_B;
		hr = spEventEditor->InsertPlayEvent(_bstr_t("edl/cmf//local/V:/default/ClipB"), section1, 0, _bstr_t(""), &editID_B);
		printf("get event properties\n");
		hr = spEventEditor->GetEventProperty (editID_B, _bstr_t("name+startpos+duration"), &varResult);
		// your code to do something with event properties

		// insert another section
		BSTR section2;
		hr = spEventEditor->InsertSection(_bstr_t(""), &section2);

		// insert several more play events and get properties back
		printf("insert play event\n");
		BSTR editID_C;
		hr = spEventEditor->InsertPlayEvent(_bstr_t("edl/cmf//local/V:/default/ClipA"), section1, 0, _bstr_t(""), &editID_C);
		printf("get event properties\n");
		hr = spEventEditor->GetEventProperty (editID_C, _bstr_t("name+startpos+duration"), &varResult);			
		// your code to do something with event properties

		printf("insert play event\n");
		BSTR editID_D;
		hr = spEventEditor->InsertPlayEvent(_bstr_t("edl/cmf//local/V:/default/ClipB"), section1, 0, _bstr_t(""), &editID_D);
		printf("get event properties\n");
		hr = spEventEditor->GetEventProperty (editID_D, _bstr_t("name+startpos+duration"), &varResult);
		// your code to do something with event properties

		// play for awhile
		printf("play for 30 seconds\n");
		hr = spPlayer->Play();
		Sleep(30000);

		// eject
		printf("eject\n");
		hr = spPlayer->Eject();

		// close channel
		printf("close channel\n\n");
		hr = spController->CloseChannel();
	}

	void GetListOfClips()
	{
		// create a mediamgr object
		IMediaMgrPtr spMediaMgr;
		hr = spAppServer->CreateMediaMgr(m_bstrAppName, &spMediaMgr);

		// enumerate the assets in the v:\default bin
		long cookie;
		long count = 0;
		BSTR bstrXmlData = NULL;	// must set this to NULL!

		spMediaMgr->EnumerateAssets(_bstr_t("V:"), _bstr_t("default"), &count, &cookie);

		// let's query for 3 properties of the clip: the name, video compression type,
		// and the length
		_bstr_t sQuery("name+videoCompressionType+lengthstr");

		// Other properties you can query:
		// "uri", "name", "assettype", "created", "modified", "attributes1", 
		// "attributes2", "videoformat", "videocompressiontype", "firsttimecodemask", 
		// "lengthstr", "maxlength", "markin", "markout", "markinstr", "markoutstr", 
		// "dropframe", "locked", "hidef"
		// for other properties search the guide for "Asset properties" 

		// Let's only get info on the first 100 clips in the bin
		int maxCount = 100;

		// Now, let's query for the results we want: 
		// query it for the properties we setup above, 
		// start with the 0th element, 
		// go up to a max of 100 clips
		// capture the actual number of clips we're getting data back from,
		// and return the results in a data object
		VARIANT varResult;
		spMediaMgr->GetResultProperty( cookie, sQuery, 0, maxCount, &count, &varResult);

		printf("Found %d clips in V:\\default\\\n", count);

		// create a safe array for the clips
		SAFEARRAY *pClipArray = varResult.parray;
		VARIANT *pVar = NULL;
		long arrClipIndex[1];

		// get the first clip
		arrClipIndex[0] = 0;
        SafeArrayPtrOfIndex(pClipArray, arrClipIndex, (void **)&pVar);

		// create a safe array for the clip's properties
		SAFEARRAY* pPropertyArray = pVar->parray;
		long arrPropertyIndex[1];

		// get the first property
		arrPropertyIndex[0] = 0;
        SafeArrayPtrOfIndex(pPropertyArray, arrPropertyIndex, (void **)&pVar);
		CString sName = pVar->bstrVal;

		// get the second property
		arrPropertyIndex[0] = 1;
        SafeArrayPtrOfIndex(pPropertyArray, arrPropertyIndex, (void **)&pVar);
		CString sVideoCompressionType = pVar->bstrVal;

		// get the third property
		arrPropertyIndex[0] = 2;
        SafeArrayPtrOfIndex(pPropertyArray, arrPropertyIndex, (void **)&pVar);
		CString sLength = pVar->bstrVal;

		printf("Clip name: %s  video compression type: %s  length: %s\n\n", sName, sVideoCompressionType, sLength);

		// etc...

		// make sure that you close the results after enumerating volumes, bins, or assets.
		// if you don't, you will run out of database selectors!
		spMediaMgr->CloseResults(cookie);
	}

	void GetVersion()
	{
		BSTR version;
		spAppServerMgrProxy->GetMSeriesIdInfo( IdComp_SW_VER, &version);
		CString sVersion = version;
		printf("K2 Software Version: %s\n", sVersion);
	}

	void Cleanup()
	{
		if ( NULL != spAppServer )
		{
			printf("close AppServer connection\n");
			spAppServer->CloseConnection();
			spAppServer = NULL;
		}

		if ( NULL != spAppServerMgrProxy )
		{
			printf("disconnect AppServerMgrProxy\n");
			VARIANT_BOOL retVal;
			spAppServerMgrProxy->Disconnect(&retVal);
			spAppServerMgrProxy = NULL;
		}
	}
};

int main(int argc, char* argv[])
{
	try
	{
		// make sure that a hostname cmdline arg was passed in
		if (argc != 2)
		{
			OutputDebugString("Usage: K2Client.exe <hostname>\n");
			printf("Usage: K2Client.exe <hostname>\n");
			exit(-1);
		}

		// define an input string
		char* input = (char*) malloc(100);

		K2Client* pK2Client = new K2Client(argv[0], argv[1]);
		
		// connect and create an AppServer
		if ( pK2Client->Connect() && pK2Client->CreateAppServer() )
		{
			// crude menu system
			while( true )
			{
				// print the menu
				printf("\n [ E] Exit\n");
				printf(" [ 1] GetSystemStatus\n");
				printf(" [ 2] GetChannelOwner\n");
				printf(" [ 3] LogTestMessage\n");
				printf(" [ 4] PlayAClip\n");
				printf(" [ 5] RecordAClip\n");
				printf(" [ 6] TransferClip\n");
				printf(" [ 7] GetChannelProperties\n");
				printf(" [ 8] GetAssetProperties\n");
				printf(" [ 9] GetXmlChanges\n");
				printf(" [10] CreatePlaylist\n");
				printf(" [11] GetListOfClips\n");
				printf(" [12] GetVersion\n");

				printf("\nPlease select a menu option: ");

				// zero input's chars and getting input string
				memset(input, 0, 100);
				fgets(input, 100, stdin);
				
				// get rid of the end line feed character
				input[ strlen(input) - 1 ] = 0;

				// compare input to known options, call appropriate functions
				if ( 0 == strcmp("E", input) || 0 == strcmp("e", input) )
					break;
				else if ( 0 == strcmp("1", input) )
					pK2Client->CheckSystemStatus();
				else if ( 0 == strcmp("2", input) )
					pK2Client->GetChannelOwner();
				else if ( 0 == strcmp("3", input) )
					pK2Client->LogTestMessage();
				else if ( 0 == strcmp("4", input) )
					pK2Client->PlayClip();
				else if ( 0 == strcmp("5", input) )
					pK2Client->RecordClip();
				else if ( 0 == strcmp("6", input) )
					pK2Client->TransferClip();
				else if ( 0 == strcmp("7", input) )
					pK2Client->GetChanStatusProperties();
				else if ( 0 == strcmp("8", input) )
					pK2Client->GetAssetProperties();
				else if ( 0 == strcmp("9", input) )
					pK2Client->GetXmlChanges();
				else if ( 0 == strcmp("10", input) )
					pK2Client->CreatePlaylist();
				else if ( 0 == strcmp("11", input) )
					pK2Client->GetListOfClips();
				else if ( 0 == strcmp("12", input) )
					pK2Client->GetVersion();
				else
					printf("\nInvalid option.\n");
			}

			// cleanup
			pK2Client->Cleanup();
			delete pK2Client;
		}
	}
	catch(_com_error &err)
	{
		printf("\nEXCEPTION - com error caught\n");
		HRESULT hr = err.Error();

		// hr = 0x80045001: StolenChannelException
		// hr = 0x80045002: ChannelInUseException		
		// hr = 0x80004005: connection refused - usually server not up and running

		printf("hresult = 0x%x\n", hr);

		mscorlib::_ExceptionPtr Exception = err.ErrorInfo();
		if(Exception)
		{
			char *bstrMessage = new char[Exception->ToString.length() + MAX_PATH];
			sprintf(bstrMessage, "%s\nPress any key to terminate...\n",
			(char*)Exception->ToString);
			printf(bstrMessage);
		}
	}
	catch( CException* e )
	{
		printf("\nEXCEPTION - exception caught\n");
		e->ReportError();
	}
	catch(...)
	{
		printf("\nEXCEPTION - unknown exception caught\n");
	}

	return 0;
}